'use strict';

function _toArray(arr) { return Array.isArray(arr) ? arr : Array.from(arr); }

var _require = require('immutable'),
    List = _require.List;

var Change = require('./Change');

/**
 * Longest common subsequence
 *
 * @param  {Array} xs, ys
 * @param  {Array} head
 * @return {Array}
 */
function lcs(xs, ys, isVariant) {
    if (xs.length > 0 && ys.length > 0) {
        var _xs = _toArray(xs),
            xe = _xs[0],
            xb = _xs.slice(1);

        var _ys = _toArray(ys),
            ye = _ys[0],
            yb = _ys.slice(1);

        if (isVariant(xe, ye)) {
            return [{
                original: xe,
                modified: ye
            }].concat(lcs(xb, yb, isVariant));
        }

        var a = lcs(xs, yb, isVariant);
        var b = lcs(xb, ys, isVariant);
        return a.length > b.length ? a : b;
    }

    return [];
}

/**
 * Diff two tree of items.
 * @param  {List}  original
 * @param  {List}  modified
 * @param  {Function} isVariant(a, b): Return true if a and b are modifiable items
 * @param  {Function} isEqual(a, b): Return true if a and b are stricly equal
 * @param  {Function} getChildren: Return children of an item
 * @return {List<Change>}
 */
function diffTree(original, modified, isVariant, isEqual, getChildren) {
    original = List(original).toArray();
    modified = List(modified).toArray();
    var subsequence = lcs(original, modified, isVariant);

    var result = [];
    var sfirst = void 0,
        mfirst = void 0,
        ofirst = void 0;

    while (subsequence.length > 0) {
        var _subsequence = subsequence;

        var _subsequence2 = _toArray(_subsequence);

        sfirst = _subsequence2[0];
        subsequence = _subsequence2.slice(1);


        while (modified.length > 0) {
            var _modified = modified;

            var _modified2 = _toArray(_modified);

            mfirst = _modified2[0];
            modified = _modified2.slice(1);

            if (isEqual(mfirst, sfirst.modified)) {
                break;
            }

            result.push(Change.createAddition(mfirst));
        }

        while (original.length > 0) {
            var _original = original;

            var _original2 = _toArray(_original);

            ofirst = _original2[0];
            original = _original2.slice(1);

            if (isEqual(ofirst, sfirst.original)) {
                break;
            }

            result.push(Change.createRemoval(ofirst));
        }

        var childrenChanges = diffTree(getChildren(sfirst.original), getChildren(sfirst.modified), isVariant, isEqual, getChildren);
        var areChildrenIdentical = childrenChanges.every(function (change) {
            return change.type == Change.TYPES.IDENTICAL;
        });

        if (areChildrenIdentical && isEqual(sfirst.original, sfirst.modified)) {
            result.push(Change.createIdentity(sfirst.original));
        } else {
            result.push(Change.createUpdate(sfirst.original, sfirst.modified, childrenChanges));
        }
    }

    while (modified.length > 0) {
        var _modified3 = modified;

        var _modified4 = _toArray(_modified3);

        mfirst = _modified4[0];
        modified = _modified4.slice(1);

        result.push(Change.createAddition(mfirst));
    }

    while (original.length > 0) {
        var _original3 = original;

        var _original4 = _toArray(_original3);

        ofirst = _original4[0];
        original = _original4.slice(1);


        result.push(Change.createRemoval(ofirst));
    }

    return List(result);
}

module.exports = diffTree;